@@ -83,25 +83,17 @@ def _compute_score(
     return scores
 
 
-@numba.njit(
-    types.int32[:](
-        types.int32,
-        types.int32,
-        types.int32[:, :],
-        types.int32[:],
-        types.float32[:],
-        types.int64,
-        types.boolean,
-    )
-)
+@numba.njit(fastmath=True)
 def randint_constrained(
     n: np.int32,
     k: np.int32,
     con_values: NDArray[np.int32],
     con_indices: NDArray[np.int32],
     p: NDArray[np.float32] = np.zeros(0, dtype=np.float32),
-    seed: np.int64 = 0,
+    seed: np.int64 = np.int64(0),
     eager: bool = False,
+    k_context: np.int32 = np.int32(-1),
+    i_forbidden: NDArray[np.int32] = np.empty(0, dtype=np.int32),
 ) -> NDArray[np.int32]:
     """
     Generate `k` unique random integers from the range `[0, n)` while satisfying given constraints.
@@ -117,9 +109,16 @@ def randint_constrained(
       compared to the older pure-Python implementation.
 
     * `con_values` & `con_indices` can be obtained by using the `to_numpy`
-       method of the [Constraints][max_div.constraints.constraints.Constraints] class.
+       method of the `Constraints` class.
 
-    *  For benchmark results, see [here](../../../../benchmarks/randint_constrained.md)
+    *  For benchmark results, see [here](../../../../benchmarks/internal/bm_randint_constrained.md)
+
+    PRIORITIES:
+
+    1) Provide exactly 'k' unique samples (no replacement)
+    2) if provided, don't generate samples from i_forbidden   (can be used to indicate already sampled values)
+    3) satisfy constraints
+    4) if p is provided, don't sample from integers with p=0
 
     :param n: range to sample from [0, n)
     :param k: number of unique samples to draw (no replacement)
@@ -130,11 +129,36 @@ def randint_constrained(
     :param eager: if True, the algorithm will try to satisfy as many constraints as early as possible; in some cases
                   increasing the probability of finding a feasible solution, albeit at the cost of sampling diversity
                   and adherence to the provided p-values.
+    :param k_context: (int, default=-1) number of total samples - in the bigger context - we want to sample in order to
+                        satisfy the constraints.  This informs the algorithm about the urgency of fulfilling
+                        constraints, giving it potentially more liberty to pick from a wider range of samples and with
+                        potentially higher p-values.
+
+                      Two cases:
+                        a) not provided or <=k:  the algorithm assumes k_context = k
+                        b) provided and >k:      the algorithm knows that more samples will be drawn later.
+
+    :param i_forbidden: (optional) 1D array of integers in `[0, n)` that must not be sampled
+
     :return: array of samples
     """
+
+    # --- parameter validation ----------------------------
+    n_forbidden = i_forbidden.shape[0]
+    if k > (n - n_forbidden):
+        if n_forbidden:
+            raise ValueError(
+                f"Cannot sample {k} unique integers from [0, {n}) when {n_forbidden} integers are forbidden."
+                f"  ({k} > {n}-{n_forbidden})"
+            )
+        else:
+            raise ValueError(f"Cannot sample {k} unique integers from [0, {n}). ({k} > {n})")
+
     # --- initialize --------------------------------------
+    if k_context < k:
+        k_context = k
     samples = np.empty(k, dtype=np.int32)
-    k_remaining = k
+    k_remaining = k_context
     m = con_values.shape[0]
 
     # Make a copy of con_values to track current min/max counts
@@ -166,11 +190,15 @@ def randint_constrained(
                 p_aug[i] += p_delta
 
     # --- sample ------------------------------------------
-    while k_remaining > 0:
+    for _ in range(k):
         # --- score & thresholds ----------------
 
         # Get already sampled integers
-        already_sampled = samples[:sample_idx]
+        # (we include i_forbidden, since they're excluded from sampling, with equal priority as already sampled values)
+        if n_forbidden:
+            already_sampled = np.concatenate((samples[:sample_idx], i_forbidden))
+        else:
+            already_sampled = samples[:sample_idx]
 
         # determine how much each integer would help us satisfy min_count constraints
         score = _compute_score(n, con_values_working, con_indices, already_sampled, True)
